@@ -3,8 +3,9 @@
 #
 
 import numpy as np
+import copy
 import torch
-
+from typing import Tuple, Union
 import gym
 from bbrl.agents.agent import Agent
 
@@ -124,7 +125,7 @@ class GymAgent(Agent):
         self.cumulated_reward = {}
         self.last_frame = {}
 
-    def _common_reset(self, k, save_render):
+    def _common_reset(self, k, save_render, render):
         env = self.envs[k]
         self.cumulated_reward[k] = 0.0
         o = env.reset()
@@ -137,6 +138,8 @@ class GymAgent(Agent):
         if save_render:
             image = env.render(mode="image").unsqueeze(0)
             observation["rendering"] = image
+        elif render:
+            env.render(mode="human")
 
         self.finished[k] = False
         self.truncated[k] = False
@@ -152,12 +155,12 @@ class GymAgent(Agent):
         }
         return _torch_type(ret), observation
 
-    def _reset(self, k, save_render):
-        ret, observation = self._common_reset(k, save_render)
+    def _reset(self, k, save_render, render):
+        ret, observation = self._common_reset(k, save_render, render)
         self.last_frame[k] = observation
         return ret
 
-    def _make_step(self, env, action, k, save_render):
+    def _make_step(self, env, action, k, save_render, render):
         action = _convert_action(action)
 
         obs, reward, done, info = env.step(action)
@@ -174,6 +177,8 @@ class GymAgent(Agent):
         if save_render:
             image = env.render(mode="image").unsqueeze(0)
             observation["rendering"] = image
+        elif render:
+            env.render(mode="human")
         ret = {
             **observation,
             "done": torch.tensor([done]),
@@ -184,7 +189,7 @@ class GymAgent(Agent):
         }
         return _torch_type(ret), done, truncated, observation
 
-    def _step(self, k, action, save_render):
+    def _step(self, k, action, save_render, render):
         if self.finished[k]:
             assert k in self.last_frame
             return {
@@ -197,7 +202,7 @@ class GymAgent(Agent):
             }
         self.timestep[k] += 1
         ret, done, truncated, observation = self._make_step(
-            self.envs[k], action, k, save_render
+            self.envs[k], action, k, save_render, render
         )
 
         self.last_frame[k] = observation
@@ -211,7 +216,7 @@ class GymAgent(Agent):
         for k in observations:
             self.set((self.output + k, t), observations[k].to(self.ghost_params.device))
 
-    def forward(self, t=0, save_render=False, **kwargs):
+    def forward(self, t=0, save_render=False, render=False, **kwargs):
         """Do one step by reading the `action` at t-1
         If t==0, environments are reset
         If save_render is True, then the output of env.render(mode="image") is written as env/rendering
@@ -221,7 +226,7 @@ class GymAgent(Agent):
             self.timestep = torch.tensor([0 for _ in self.envs])
             observations = []
             for k, e in enumerate(self.envs):
-                obs = self._reset(k, save_render)
+                obs = self._reset(k, save_render, render)
                 observations.append(obs)
             observations = _torch_cat_dict(observations)
             for k in observations:
@@ -234,7 +239,7 @@ class GymAgent(Agent):
             assert action.size()[0] == self.n_envs, "Incompatible number of envs"
             observations = []
             for k, e in enumerate(self.envs):
-                obs = self._step(k, action[k], save_render)
+                obs = self._step(k, action[k], save_render, render)
                 observations.append(obs)
             self.set_obs(observations, t)
 
@@ -280,12 +285,11 @@ class AutoResetGymAgent(GymAgent):
 
         Args:
             make_env_fn ([function that returns a gym.Env]): The function to create a single gym environments
-            make_env_args (dict): The arguments of the function that creates a gym.Env
-            n_envs ([int]): The number of environments to create.
+            make_env_args (dict): [The arguments of the function that creates a gym.Env]
+            n_envs ([int]): [The number of environments to create].
+            seed (int, optional): [the seed used to initialize the environment].
             action_string (str, optional): [the name of the action variable in the workspace]. Defaults to "action".
             output (str, optional): [the output prefix of the environment]. Defaults to "env/".
-            seed (int): the seed used to initialize the environment
-            and each environment will have its own seed]. Defaults to True.
         """
         super().__init__(
             make_env_fn=make_env_fn,
@@ -297,20 +301,20 @@ class AutoResetGymAgent(GymAgent):
         )
         self.is_running = [False for _ in range(self.n_envs)]
 
-    def _reset(self, k, save_render):
+    def _reset(self, k, save_render, render):
         self.is_running[k] = True
-        ret, _ = self._common_reset(k, save_render)
+        ret, _ = self._common_reset(k, save_render, render)
         return ret
 
-    def _step(self, k, action, save_render):
+    def _step(self, k, action, save_render, render):
         self.timestep[k] += 1
-        ret, done, truncated, _ = self._make_step(self.envs[k], action, k, save_render)
+        ret, done, truncated, _ = self._make_step(self.envs[k], action, k, save_render, render)
         if done:
             self.is_running[k] = False
             self.truncated[k] = truncated
         return ret
 
-    def forward(self, t=0, save_render=False, **kwargs):
+    def forward(self, t=0, save_render=False, render=False, **kwargs):
         """
         Perform one step by reading the `action`
         """
@@ -318,12 +322,12 @@ class AutoResetGymAgent(GymAgent):
         observations = []
         for k, env in enumerate(self.envs):
             if not self.is_running[k] or t == 0:
-                observations.append(self._reset(k, save_render))
+                observations.append(self._reset(k, save_render, render))
             else:
                 assert t > 0
                 action = self.get((self.input, t - 1))
                 assert action.size()[0] == self.n_envs, "Incompatible number of envs"
-                observations.append(self._step(k, action[k], save_render))
+                observations.append(self._step(k, action[k], save_render, render))
 
         self.set_obs(observations, t)
 
@@ -348,3 +352,219 @@ class NoAutoResetGymAgent(GymAgent):
             action_string=action_string,
             output=output,
         )
+
+# --------------------------- part added from the code of Folco Bertini and Nikola Matevski --------------------
+
+class RunningMeanStd:
+    def __init__(self, epsilon: float = 1e-4, shape: Tuple[int, ...] = ()):
+        """
+        Calulates the running mean and std of a data stream
+        https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
+        :param epsilon: helps with arithmetic issues
+        :param shape: the shape of the data stream's output
+        """
+        self.mean = torch.zeros(shape, dtype=torch.float32)
+        self.var = torch.ones(shape, dtype=torch.float32)
+        self.count = epsilon
+        self.steps = 0
+        self.update_threshold = 4
+        self.buffer = None
+
+    def copy(self) -> "RunningMeanStd":
+        """
+        :return: Return a copy of the current object.
+        """
+        new_object = RunningMeanStd(shape=self.mean.shape)
+        new_object.mean = copy.deepcopy(self.mean)
+        new_object.var = copy.deepcopy(self.var)
+        new_object.count = float(self.count)
+        return new_object
+
+    def combine(self, other: "RunningMeanStd") -> None:
+        """
+        Combine stats from another ``RunningMeanStd`` object.
+        :param other: The other object to combine with.
+        """
+        self.update_from_moments(other.mean, other.var, other.count)
+
+    def update(self, arr: torch.Tensor) -> None:
+        self.steps += 1
+        if self.buffer is None:
+            self.buffer = copy.deepcopy(arr)
+        else:
+            self.buffer = torch.concat((self.buffer, copy.deepcopy(arr)), dim=0)
+        if self.steps > self.update_threshold:
+            batch_mean = torch.mean(self.buffer, dim=0)
+            batch_var = torch.var(self.buffer, dim=0)
+            batch_count = self.buffer.size(0)
+            self.steps = 0
+            self.buffer = None
+            self.update_from_moments(batch_mean, batch_var, batch_count)
+
+    def update_from_moments(
+        self,
+        batch_mean: torch.Tensor,
+        batch_var: torch.Tensor,
+        batch_count: Union[int, float],
+    ) -> None:
+        delta = batch_mean - self.mean
+        tot_count = self.count + batch_count
+
+        new_mean = self.mean + delta * batch_count / tot_count
+        m_a = self.var * self.count
+        m_b = batch_var * batch_count
+        m_2 = (
+            m_a
+            + m_b
+            + torch.square(delta)
+            * self.count
+            * batch_count
+            / (self.count + batch_count)
+        )
+        new_var = m_2 / (self.count + batch_count)
+
+        new_count = batch_count + self.count
+
+        self.mean = new_mean
+        self.var = new_var
+        self.count = new_count
+
+
+# not checked
+class NormalizedNoAutoResetGymAgent(NoAutoResetGymAgent):
+    """The same as AutoResetGymAgent, but normalizes observations"""
+
+    def __init__(
+        self,
+        make_env_fn=None,
+        make_env_args={},
+        n_envs=None,
+        seed=None,
+        action_string="action",
+        output="env/",
+        obs_rms=None,
+    ):
+        super().__init__(
+            make_env_fn=make_env_fn,
+            make_env_args=make_env_args,
+            n_envs=n_envs,
+            seed=seed,
+            action_string=action_string,
+            output=output,
+        )
+        self.clip_obs = 10.0
+        self.epsilon = 1e-08
+        self.obs_rms = obs_rms
+
+    def normalize_obs(self, obs: np.ndarray) -> np.ndarray:
+        """
+        Helper to normalize observation.
+        :param obs:
+        :return: normalized observation
+        """
+        obs_ = copy.deepcopy(obs)
+        obs_rms = self.obs_rms
+        return np.clip(
+            (obs_ - obs_rms.mean) / np.sqrt(obs_rms.var + self.epsilon),
+            -self.clip_obs,
+            self.clip_obs,
+        )
+
+    def _reset(self, k, save_render, render):
+        full_obs, observation = self._common_reset(k, save_render, render)
+        self.last_frame[k] = observation
+
+        full_obs["env_obs"] = self.normalize_obs(full_obs["env_obs"])
+
+        return full_obs
+
+    def _step(self, k, action, save_render, render):
+        if self.finished[k]:
+            assert k in self.last_frame
+            rew = _torch_type({"reward": torch.tensor([0.0]).float()})
+            return (
+                {
+                    **self.last_frame[k],
+                    "done": torch.tensor([True]),
+                    "truncated": torch.tensor([self.truncated[k]]),
+                    "cumulated_reward": torch.tensor(
+                        [self.cumulated_reward[k]]
+                    ).float(),
+                    "timestep": torch.tensor([self.timestep[k]]),
+                },
+                rew,
+            )
+        self.timestep[k] += 1
+        ret, done, truncated, observation = self._make_step(
+            self.envs[k], action, k, save_render, render
+        )
+
+        self.last_frame[k] = observation
+        if done:
+            self.finished[k] = True
+            self.truncated[k] = truncated
+
+        ret["env_obs"] = self.normalize_obs(ret["env_obs"])
+        return ret
+
+# not checked
+class NormalizedAutoResetGymAgent(AutoResetGymAgent):
+    """The same as AutoResetGymAgent, but normalizes observations"""
+
+    def __init__(
+        self,
+        make_env_fn=None,
+        make_env_args={},
+        n_envs=None,
+        seed=None,
+        action_string="action",
+        output="env/",
+    ):
+        super().__init__(
+            make_env_fn=make_env_fn,
+            make_env_args=make_env_args,
+            n_envs=n_envs,
+            seed=seed,
+            action_string=action_string,
+            output=output,
+        )
+        self.clip_obs = 10.0
+        self.epsilon = 1e-08
+        self.obs_rms = RunningMeanStd(shape=self.observation_space.shape)
+
+    def normalize_obs(self, obs: np.ndarray) -> np.ndarray:
+        """
+        Helper to normalize observation.
+        :param obs:
+        :return: normalized observation
+        """
+        obs_ = copy.deepcopy(obs)
+        obs_rms = self.obs_rms
+
+        return np.clip(
+            (obs_ - obs_rms.mean) / np.sqrt(obs_rms.var + self.epsilon),
+            -self.clip_obs,
+            self.clip_obs,
+        )
+
+    def _reset(self, k, save_render, render):
+        self.is_running[k] = True
+        full_obs, _ = self._common_reset(k, save_render, render)
+
+        self.obs_rms.update(full_obs["env_obs"])
+        full_obs["env_obs"] = self.normalize_obs(full_obs["env_obs"])
+        return full_obs
+
+    def _step(self, k, action, save_render, render):
+        self.timestep[k] += 1
+        ret, done, truncated, _ = self._make_step(
+            self.envs[k], action, k, save_render, render
+        )
+        if done:
+            self.is_running[k] = False
+            self.truncated[k] = truncated
+
+        self.obs_rms.update(ret["env_obs"])
+        ret["env_obs"] = self.normalize_obs(ret["env_obs"])
+
+        return ret
